/*
    libview64 image C64 decoding library
    (c) 1995-2013 Soci/Singular (soci at c64.rulez.org)

    $Id: libview64.h 88 2016-04-24 16:42:35Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
#ifndef _LIBVIEW64_H
#define _LIBVIEW64_H

#include <stdlib.h>

typedef struct {
    unsigned int  flags; /* screen flags */
    int w, h;            /* width, height */
    unsigned char *b;    /* border left and right color */
    unsigned char br;    /* border top/bottom color */
    unsigned char n;     /* number of screens 1 / 2 */
    unsigned char *d[2]; /* screen datas w * h bytes */
} view64_screen;

typedef enum {
    VIEW64_SCREEN_MCI        = 0x01, /* fake deinterlacing possible */
    VIEW64_SCREEN_BORDERLESS = 0x02, /* image has no border */
} view64_screen_flags;

typedef enum {
    VIEW64_UNKNOWN, VIEW64_GODOT, VIEW64_VBM, VIEW64_FLI, VIEW64_EMC_EDITOR,
    VIEW64_FLI_EDITOR, VIEW64_FLIGRAPH, VIEW64_FLIP, VIEW64_FLIPROFI,
    VIEW64_FLASH_FLI, VIEW64_HCB, VIEW64_SUPER_HIRES_FLI, VIEW64_XFLI,
    VIEW64_UFLI, VIEW64_MUFLI, VIEW64_NUFLI, VIEW64_MUIFLI, VIEW64_UIFLI,
    VIEW64_SHFXL, VIEW64_SUPER_HIRES_INTERLACE_FLI,
    VIEW64_SUPER_HIRES_INTERLACE, VIEW64_EXTEND_SUPER_HIRES_INTERLACE,
    VIEW64_SUPER_HIRES_STUDIO, VIEW64_SUPER_HIRES, VIEW64_SUPER_HIRES_2,
    VIEW64_SUPER_HIRES_II, VIEW64_SUPER_HIRES_I, VIEW64_FUCKPAINT,
    VIEW64_TRUEPAINT, VIEW64_GUNPAINT, VIEW64_WIGMORE_ARTIST,
    VIEW64_ART_STUDIO_MULTI, VIEW64_ZOOMATIC, VIEW64_DRAZPAINT,
    VIEW64_DRAZLACE, VIEW64_FUNPAINT, VIEW64_PIXELPERFECT, VIEW64_FACEPAINT,
    VIEW64_ART_STUDIO_HIRES, VIEW64_FONT, VIEW64_HIRES_INTERLACE,
    VIEW64_HIRESLACE_EDITOR, VIEW64_DOODLE, VIEW64_IMAGE_SYSTEM_HIRES,
    VIEW64_HIEDDI, VIEW64_HIRES_EDITOR, VIEW64_IMAGE_SYSTEM_MULTI,
    VIEW64_VIDCOM64, VIEW64_CDUPAINT, VIEW64_KOALA_PAINTER,
    VIEW64_BLAZING_PADDLES, VIEW64_ARTIST64, VIEW64_CHEESE_PAINT,
    VIEW64_GARFIELD, VIEW64_SARACEN_PAINT, VIEW64_DOLPHIN,
    VIEW64_MICRO_ILLUSTRATOR, VIEW64_AFLI, VIEW64_FLIMATIC,
    VIEW64_HIRES_MANAGER, VIEW64_HIRES_FLI_DESIGN, VIEW64_ECI, VIEW64_AMICA,
    VIEW64_PAINT_MAGIC, VIEW64_BITMAP_EDITOR, VIEW64_RAINBOW_PAINTER,
    VIEW64_RUN_PAINT_MULTI, VIEW64_RUN_PAINT_HIRES, VIEW64_RUN_PAINT_MONO,
    VIEW64_INTERPAINT_HIRES, VIEW64_INTERPAINT_MULTI, VIEW64_GIGA_PAINT_HIRES,
    VIEW64_GIGA_PAINT_MULTI, VIEW64_FLI_DESIGNER, VIEW64_INTERLACE_SUPER_HIRES,
    VIEW64_PICASSO64, VIEW64_LOGO_PAINTER_V3, VIEW64_BIGFLI,
    VIEW64_FLYING_COLORS, VIEW64_CENTAURI_LOGO_EDITOR,
    VIEW64_INTERLACED_LOGO_EDITOR, VIEW64_ART_STUDIO_MULTI_WINDOW,
    VIEW64_PAINT_MAGIC_PLUS, VIEW64_CHARPAD, VIEW64_SPRITEPAD, VIEW64_INVALID
} view64_format;

#ifdef __cplusplus
extern "C" {
#endif

const char *view64_format_name(view64_format);
view64_format view64_format_detect(const char *, const unsigned char *, size_t);
view64_screen *view64_decode(view64_format, const unsigned char *, size_t, unsigned int);
void view64_free_screen(view64_screen *);

#ifdef __cplusplus
}
#endif

#endif

