/*
    View64 C64 image viewer
    (c) 1995-2013 Soci/Singular (soci at c64.rulez.org)

    $Id: main.c 70 2015-10-31 17:23:02Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#ifdef _WIN32
#include <windows.h>
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <locale.h>
#include <wchar.h>
#include "main.h"

#ifdef __DJGPP__
static size_t mbrtowc(wchar_t *wc, const char *s, size_t n, mbstate_t *ps) {
    if (!*s) return 0;
    *wc = *s;
    return 1;
}

static size_t wcrtomb(char *s, wchar_t wc, mbstate_t *ps) {
    *s = wc;
    return 1;
}
#endif

extern int main2(int, char *[]);

static unsigned int utf8in(const unsigned char *c, unsigned int *out) { /* only for internal use with validated utf-8! */
    unsigned int ch;
    int i, j;
    ch = c[0];

    if (ch < 0xe0) {
        ch ^= 0xc0;i = 2;
    } else if (ch < 0xf0) {
        ch ^= 0xe0;i = 3;
    } else if (ch < 0xf8) {
        ch ^= 0xf0;i = 4;
    } else if (ch < 0xfc) {
        ch ^= 0xf8;i = 5;
    } else {
        ch ^= 0xfc;i = 6;
    }

    for (j = 1;c[j] && j < i; j++) {
        ch = (ch << 6) ^ c[j] ^ 0x80;
    }
    *out = ch;
    return i;
}

FILE *fopen2(const char *name, const char *mode)
{
    FILE *f;
#ifdef _WIN32
    wchar_t *wname, *c2, wmode[3];
    const unsigned char *c;
    unsigned int ch;
    size_t len = strlen(name) + 1;
    if (len > SIZE_MAX / sizeof(wchar_t)) return NULL;
    wname = (wchar_t *)malloc(len * sizeof(wchar_t));
    if (!wname) return NULL;
    c2 = wname; c = (unsigned char *)name;
    while (*c) {
        ch = *c;
        if (ch & 0x80) c += utf8in(c, &ch); else c++;
        if (ch < 0x10000) *c2++ = ch;
        else if (ch < 0x110000) {
            *c2++ = (ch >> 10) + 0xd7c0;
            *c2++ = (ch & 0x3ff) | 0xdc00;
        } else *c2++ = 0xfffd;
    }
    *c2++ = 0;
    c2 = wmode; c = (unsigned char *)mode;
    while ((*c2++=(wchar_t)*c++));
    f=_wfopen(wname, wmode);
    free(wname);
#else
    size_t len = 0, max = strlen(name) + 1;
    char *newname = (char *)malloc(max);
    const unsigned char *c = (unsigned char *)name;
    unsigned int ch;
    mbstate_t ps;
    memset(&ps, 0, sizeof(mbstate_t));
    if (newname == NULL || max < 1) return NULL;
    do {
        char temp[64];
        int l;
        ch = *c;
        if (ch & 0x80) c += utf8in(c, &ch); else c++;
        l = wcrtomb(temp, (wchar_t)ch, &ps);
        if (l <= 0) l = sprintf(temp, "{$%x}", ch);
        len += l;
        if (len < (size_t)l) {free(newname); return NULL;}
        if (len > max) {
            max = len + 64;
            if (max < 64) {free(newname); return NULL;}
            newname = (char *)realloc(newname, max);
            if (!newname) return NULL;
        }
        memcpy(newname + len - l, temp, l);
    } while (ch);
    f=fopen(newname, mode);
    free(newname);
#endif
    return f;
}

static unsigned char *utf8out(unsigned int i, unsigned char *c) {
    if (i < 0x800) {
        *c++=0xc0 | (i >> 6);
        *c++=0x80 | (i & 0x3f);
	return c;
    }
    if (i < 0x10000) {
        *c++=0xe0 | (i >> 12);
        *c++=0x80 | ((i >> 6) & 0x3f);
        *c++=0x80 | (i & 0x3f);
	return c;
    }
    if (i < 0x200000) {
        *c++=0xf0 | (i >> 18);
        *c++=0x80 | ((i >> 12) & 0x3f);
        *c++=0x80 | ((i >> 6) & 0x3f);
        *c++=0x80 | (i & 0x3f);
	return c;
    }
    if (i < 0x4000000) {
        *c++=0xf8 | (i >> 24);
        *c++=0x80 | ((i >> 18) & 0x3f);
        *c++=0x80 | ((i >> 12) & 0x3f);
        *c++=0x80 | ((i >> 6) & 0x3f);
        *c++=0x80 | (i & 0x3f);
	return c;
    }
    if (i & ~0x7fffffff) return c;
    *c++=0xfc | (i >> 30);
    *c++=0x80 | ((i >> 24) & 0x3f);
    *c++=0x80 | ((i >> 18) & 0x3f);
    *c++=0x80 | ((i >> 12) & 0x3f);
    *c++=0x80 | ((i >> 6) & 0x3f);
    *c++=0x80 | (i & 0x3f);
    return c;
}

#ifdef _WIN32
extern int wmain(int, wchar_t *[]);

int wmain(int argc, wchar_t *argv2[]) {
    int i, r;
    char **argv;

    argv = (char **)malloc(sizeof(char *)*argc);
    if (!argv) return 1;
    for (i = 0; i < argc; i++) {
	unsigned int c = 0, lastchar;
	wchar_t *p = argv2[i];
	unsigned char *c2;

	while (*p) p++;
	c2 = (unsigned char *)malloc((p - argv2[i])*4/sizeof(wchar_t)+1);
	if (!c2) return 1;
	p = argv2[i];
	argv[i] = (char *)c2;

	while (*p) {
	    lastchar = c;
	    c = *p++;
	    if (c >= 0xd800 && c < 0xdc00) {
		if (lastchar < 0xd800 || lastchar >= 0xdc00) continue;
		c = 0xfffd;
	    } else if (c >= 0xdc00 && c < 0xe000) {
		if (lastchar >= 0xd800 && lastchar < 0xdc00) {
		    c ^= 0x360dc00 ^ (lastchar << 10);
		    c += 0x10000;
		} else
		    c = 0xfffd;
	    } else if (lastchar >= 0xd800 && lastchar < 0xdc00) {
		c = 0xfffd;
	    }
	    if (c && c < 0x80) *c2++ = c; else c2 = utf8out(c, c2);
	}
	*c2++ = 0;
	argv[i] = (char *)realloc(argv[i], (char *)c2 - argv[i]);
	if (!argv[i]) return 1;
    }
    r = main2(argc, argv);

    for (i = 0; i < argc; i++) free(argv[i]);
    free(argv);
    return r;
}
#else
int main(int argc, char *argv[]) {
    int i, r;
    char **uargv;

    setlocale(LC_ALL, "");
    setlocale(LC_NUMERIC, "C");

    uargv = (char **)malloc(sizeof(char *)*argc);
    if (!uargv) return 1;
    for (i = 0; i < argc; i++) {
        const char *s = argv[i];
        mbstate_t ps;
        unsigned char *p;
        size_t n = strlen(s), j = 0;
        size_t len = n + 64;
        unsigned char *data = (unsigned char *)malloc(len);
        if (!data || len < 64) return 1;

        memset(&ps, 0, sizeof(ps));
        p = data;
        for (;;) {
            ssize_t l;
            wchar_t w;
            unsigned int ch;
            if (p + 6*6 + 1 > data + len) {
                size_t o = p - data;
                len += 1024;
                data=(unsigned char *)realloc(data, len);
                if (!data) return 1;
                p = data + o;
            }
            l = mbrtowc(&w, s + j, n - j,  &ps);
            if (l < 1) break;
            j += l;
            ch = w;
            if (ch && ch < 0x80) *p++ = ch; else p = utf8out(ch, p);
        }
        *p++ = 0;
        uargv[i] = (char *)data;
    }
    r = main2(argc, uargv);

    for (i = 0; i < argc; i++) free(uargv[i]);
    free(uargv);
    return r;
}
#endif

#ifdef __MINGW32__

#include <shellapi.h>

int main(void)
{
  LPWSTR commandLine = GetCommandLineW();
  int argcw = 0;
  LPWSTR *argvw = CommandLineToArgvW(commandLine, &argcw);
  if (!argvw)
    return 127;

  int result = wmain(argcw, argvw);
  LocalFree(argvw);
  return result;
}
#endif /* __MINGW32__ */
