/*
    tvrender for libview64
    (c) 2013 Soci/Singular (soci at c64.rulez.org)

    $Id: tvrender.c 88 2016-04-24 16:42:35Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include <string.h>
#include <stdio.h>
#include <math.h>
#include "tvrender.h"

static rgb_color pal_palette[16] = {
    { 0x0f, 0x0f, 0x0f },
    { 0xfa, 0xfa, 0xfa },
    { 0x7f, 0x27, 0x34 },
    { 0x6f, 0xc0, 0xb4 },
    { 0x87, 0x3c, 0x95 },
    { 0x56, 0x9c, 0x49 },
    { 0x3e, 0x3e, 0x8e },
    { 0xbb, 0xbb, 0x6a },
    { 0x86, 0x3d, 0x27 },
    { 0x55, 0x2c, 0x00 },
    { 0xae, 0x56, 0x63 },
    { 0x4e, 0x4e, 0x4e },
    { 0x75, 0x75, 0x75 },
    { 0x94, 0xda, 0x87 },
    { 0x75, 0x75, 0xc5 },
    { 0x9c, 0x9c, 0x9c }
};

static rgb_color pal_old_palette[16] = {
    { 0x0f, 0x0f, 0x0f },
    { 0xfa, 0xfa, 0xfa },
    { 0x6f, 0x17, 0x24 },
    { 0x8e, 0xdf, 0xd3 },
    { 0xa7, 0x5c, 0xb5 },
    { 0x56, 0x9c, 0x49 },
    { 0x3e, 0x3e, 0x8e },
    { 0xbb, 0xbb, 0x6a },
    { 0xa6, 0x5d, 0x47 },
    { 0x55, 0x2c, 0x00 },
    { 0xae, 0x56, 0x63 },
    { 0x3e, 0x3e, 0x3e },
    { 0x7d, 0x7d, 0x7d },
    { 0x94, 0xda, 0x87 },
    { 0x7d, 0x7d, 0xcd },
    { 0xbb, 0xbb, 0xbb }
};

static tvrender_image *new_tv_pixels(int w, int h, int p, int scale) {
    tvrender_image *pixels;

    pixels = (tvrender_image *)malloc(sizeof(tvrender_image));
    if (pixels == NULL) return NULL;
    pixels->pitch = p;
    pixels->h = h;
    pixels->w = w;
    pixels->scale = scale;
    pixels->colors = 0;
    pixels->palette = NULL;
    pixels->d = (unsigned char *)malloc(pixels->pitch * pixels->h);
    if (pixels->d == NULL) {
        free(pixels);
        return NULL;
    }
    return pixels;
}

void free_tv_pixels(tvrender_image *pixels) {
    free(pixels->palette);
    free(pixels->d);
    free(pixels);
}

static tvrender_image *create_pal(const view64_screen *rscreen, int scale, int frame, unsigned int vmode) {
    tvrender_image *pixels;
    const int *luma;
    const static int luma1[16] = {
        2, 32, 10, 20, 12, 16, 8, 24, 12, 8, 16, 10, 15, 24, 15, 20
    };
    const static int luma2[16] = {
        2, 32, 8, 24, 16, 16, 8, 24, 16, 8, 16, 8, 16, 24, 16, 24
    };
    const static signed char cols[32][16] = {
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 0
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1
        { -24, -4, 16, 35, 48, 53, 51, 40, 24, 4, -16, -35, -48, -53, -51, -40}, // 2
        { 22, 3, -15, -32, -44, -50, -47, -38, -22, -3, 15, 32, 44, 50, 47, 38}, // 3
        { 39, 53, 60, 57, 45, 27, 4, -18, -39, -53, -60, -57, -45, -27, -4, 18}, // 4
        { -36, -50, -56, -53, -42, -25, -4, 17, 36, 50, 56, 53, 42, 25, 4, -17}, // 5
        { 60, 57, 45, 27, 4, -18, -39, -53, -60, -57, -45, -27, -4, 18, 39, 53}, // 6
        { -61, -58, -46, -27, -4, 19, 40, 54, 61, 58, 46, 27, 4, -19, -40, -54}, // 7
        { -44, -26, -4, 18, 38, 52, 58, 56, 44, 26, 4, -18, -38, -52, -58, -56}, // 8
        { -59, -47, -28, -4, 19, 40, 55, 62, 59, 47, 28, 4, -19, -40, -55, -62}, // 9
        { -24, -4, 16, 35, 48, 53, 51, 40, 24, 4, -16, -35, -48, -53, -51, -40}, // 10
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 11
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 12
        { -36, -50, -56, -53, -42, -25, -4, 17, 36, 50, 56, 53, 42, 25, 4, -17}, // 13
        { 60, 57, 45, 27, 4, -18, -39, -53, -60, -57, -45, -27, -4, 18, 39, 53}, // 14
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 15

        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 0
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 1
        { 16, 35, 48, 53, 51, 40, 24, 4, -16, -35, -48, -53, -51, -40, -24, -4}, // 2
        { -15, -32, -44, -50, -47, -38, -22, -3, 15, 32, 44, 50, 47, 38, 22, 3}, // 3
        { -45, -27, -4, 18, 39, 53, 60, 57, 45, 27, 4, -18, -39, -53, -60, -57}, // 4
        { 42, 25, 4, -17, -36, -50, -56, -53, -42, -25, -4, 17, 36, 50, 56, 53}, // 5
        { -60, -57, -45, -27, -4, 18, 39, 53, 60, 57, 45, 27, 4, -18, -39, -53}, // 6
        { 61, 58, 46, 27, 4, -19, -40, -54, -61, -58, -46, -27, -4, 19, 40, 54}, // 7
        { 38, 52, 58, 56, 44, 26, 4, -18, -38, -52, -58, -56, -44, -26, -4, 18}, // 8
        { 55, 62, 59, 47, 28, 4, -19, -40, -55, -62, -59, -47, -28, -4, 19, 40}, // 9
        { 16, 35, 48, 53, 51, 40, 24, 4, -16, -35, -48, -53, -51, -40, -24, -4}, // 10
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 11
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // 12
        { 42, 25, 4, -17, -36, -50, -56, -53, -42, -25, -4, 17, 36, 50, 56, 53}, // 13
        { -60, -57, -45, -27, -4, 18, 39, 53, 60, 57, 45, 27, 4, -18, -39, -53}, // 14
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}  // 15
    };
    unsigned char *p, *d;
    int x, y, w, lb, n2;
    signed char *l, *ol;
    unsigned int *offs = NULL;
    d = rscreen->d[frame];
    pixels = new_tv_pixels(width_tvpicture(rscreen->w, vmode, scale), rscreen->h, (((rscreen->w + 4) * scale * 561 / 200 + 2) / 3) * 3 + 6, scale);
    w = rscreen->w + 4;
    lb = 2;
    n2 = ((((pixels->pitch / 3) - pixels->w) / 2) * 3000 - lb * scale * 2805) + (scale == 1 ? 2422 : 3366);
    n2 += ((((pixels->pitch / 3) - pixels->w) / 2 + pixels->w) * 3000 - (lb + rscreen->w) * scale * 2805) + (scale == 1 ? 1580 : 2633);
    n2 /= 2;

    l = (signed char *)malloc((pixels->pitch + 3 * scale) * 600 / 187 / scale * 9 + 14 + 16);
    if (l == NULL) return NULL;
    ol = (signed char *)malloc((pixels->pitch + 3 * scale) * 600 / 187 / scale * 9 + 14 + 16);
    if (ol == NULL) {
        free(l);
        return NULL;
    }
    for (x = 0; x < (pixels->pitch + 3 * scale) * 600 / 187 / scale * 9 + 14 + 16; x++) ol[x] = cols[rscreen->br + 16][x & 15];
    memcpy(l, ol, (pixels->pitch + 3 * scale) * 600 / 187 / scale * 9 + 14 + 16);

    if (!(vmode & VIDEO_MONOCHROME)) {
        int a, i2, j;
        int scale2 = scale * 8;
        offs = (unsigned int *)malloc(sizeof(unsigned int *) * ((pixels->pitch / 3) + 2 * scale));
        if (offs == NULL) {
            free(ol);
            free(l);
            return NULL;
        }
        for (j = i2 = a = x = 0; x < (pixels->pitch / 3) + 2 * scale; x++) {
            offs[x] = j;
            a += (600 * 3) % 187;
            if (a >= 187) {
                a -= 187;
                i2 += (600 * 3) / 187 + 1;
            } else i2 += (600 * 3) / 187;
            while (i2 >= scale2) {
                i2 -= scale2;
                j += 8;
            }
        }
    }

    luma = (vmode & VIDEO_OLD_LUMA) ? luma2 : luma1;
    memset(pixels->d, 0, pixels->pitch * pixels->h);
    for (y = 0; y < rscreen->h; y++) {
        int i, n = n2, ci = 0;
        int accu = 0;
        signed char *l2;
        p = pixels->d + y * pixels->pitch;
        l2 = l;
        l = ol;
        ol = l2;
        {
            int l;
            const signed char *c;
            int cb = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
            c = cols[cb + 16 * (y & 1)];
            for (l = 0; l < 14; l++) {
                l2[ci] = c[ci];
                ci++;
            }
        }
        for (x = -lb; x < w - lb; x++) {
            int l, ll;
            const signed char *c;
            if (x < 0) {
                int cb = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
                if (rscreen->flags & VIEW64_SCREEN_BORDERLESS) cb = *d;
                l = luma[cb];
                c = cols[cb + 16 * (y & 1)];
            } else if (x >= rscreen->w) {
                int cb = rscreen->b ? rscreen->b[y * 2 + 1] : rscreen->br;
                if (rscreen->flags & VIEW64_SCREEN_BORDERLESS) cb = d[-1];
                l = luma[cb];
                c = cols[cb + 16 * (y & 1)];
            } else {
                l = luma[*d];
                c = cols[*d + 16 * (y & 1)];
                d++;
            }
            accu += (1000 - n) * l;
            *p++ = accu / 128;
            n += 2805 * scale - 1000;
            ll = 1000 * l / 128;
            while (n >= 1000) {
                *p++ = ll;
                n -= 1000;
            }
            accu = n * l;
            if (c[0]) {
                for (l = 0; l < 9; l++) {
                    l2[ci] = c[ci & 15];
                    ci++;
                }
            } else {
                memset(l2 + ci, 0, 9);
                ci += 9;
            }
        }
        p = pixels->d + y * pixels->pitch;
        if (scale == 1) {
            int a, b;
            a = b = p[2];
            for (x = 0; x < 4; x++) a += p[x];
            for (x = 0; x < pixels->pitch - 4; x++) {
                a += p[x+4] - b;
                b = p[x];
                p[x] = a / 5;
            }
        } else {
            int a, b;
            a = b = p[2];
            for (x = 0; x < 6; x++) a += p[x];
            for (x = 0; x < pixels->pitch - 6; x++) {
                a += p[x+6] - b;
                b = p[x];
                p[x] = a / 7;
            }
        }
        if (!(vmode & VIDEO_MONOCHROME)) {
            unsigned int *offs1 = offs, *offs2 = offs;
            int r = 0, g = 0, b = 0;
            if (y & 1) {
                for (x = 0; x < 2 * scale; x++) {
                    int j = *offs2++;
                    int e = l[j + 4] + ol[j + 4];
                    if (j & 8) {
                        r -= e;
                        g += e;
                        b += l[j + 8] - ol[j + 8];
                    } else {
                        r += e;
                        g -= e;
                        b -= l[j + 8] - ol[j + 8];
                    }
                }
                for (x = 0; x < pixels->pitch; x += 3) {
                    int j;
                    int v;
                    int j2 = *offs1++;
                    int e = l[j2 + 4] + ol[j2 + 4];
                    if (j2 & 8) {
                        r += e;
                        g -= e;
                        b -= l[j2 + 8] - ol[j2 + 8];
                    } else {
                        r -= e;
                        g += e;
                        b += l[j2 + 8] - ol[j2 + 8];
                    }
                    j = *offs2++;
                    e = l[j + 4] + ol[j + 4];
                    if (j & 8) {
                        r -= e;
                        g += e;
                        b += l[j + 8] - ol[j + 8];
                    } else {
                        r += e;
                        g -= e;
                        b -= l[j + 8] - ol[j + 8];
                    }
                    v = p[x];
                    v += r / 4 / scale;
                    p[x] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                    v = p[x+1];
                    v += g / 5 / scale;
                    p[x+1] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                    v = p[x+2];
                    v += b / 3 / scale;
                    p[x+2] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                }
            } else {
                for (x = 0; x < 2 * scale; x++) {
                    int j = *offs2++;
                    int e = l[j + 4] + ol[j + 4];
                    if (j & 8) {
                        r -= e;
                        g += e;
                        b -= l[j + 8] - ol[j + 8];
                    } else {
                        r += e;
                        g -= e;
                        b += l[j + 8] - ol[j + 8];
                    }
                }
                for (x = 0; x < pixels->pitch; x += 3) {
                    int j;
                    int v;
                    int j2 = *offs1++;
                    int e = l[j2 + 4] + ol[j2 + 4];
                    if (j2 & 8) {
                        r += e;
                        g -= e;
                        b += l[j2 + 8] - ol[j2 + 8];
                    } else {
                        r -= e;
                        g += e;
                        b -= l[j2 + 8] - ol[j2 + 8];
                    }
                    j = *offs2++;
                    e = l[j + 4] + ol[j + 4];
                    if (j & 8) {
                        r -= e;
                        g += e;
                        b -= l[j + 8] - ol[j + 8];
                    } else {
                        r += e;
                        g -= e;
                        b += l[j + 8] - ol[j + 8];
                    }
                    v = p[x];
                    v += r / 4 / scale;
                    p[x] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                    v = p[x+1];
                    v += g / 5 / scale;
                    p[x+1] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                    v = p[x+2];
                    v += b / 3 / scale;
                    p[x+2] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                }
            }
        }
        if (vmode & VIDEO_CHROMA_NOISE) {
            int a, j;

            for (j = a = i = x = 0; x < pixels->pitch; x++) {
                int v = p[x] << 4;
                v += l2[j + 4] + 8;
                p[x] = (v <= 0) ? 0 : (v >= (255 << 4)) ? 255 : (v >> 4);
                a += 600 % 187;
                if (a >= 187) {
                    a -= 187;
                    i += (600 / 187) + 1;
                } else i += 600 / 187;
                while (i >= scale) {
                    i -= scale; 
                    j++;
                }
            }
        }
    }
    free(offs);
    free(ol);
    free(l);
    return pixels;
}

static tvrender_image *create_ntsc(const view64_screen *rscreen, int scale, int frame, unsigned int vmode) {
    tvrender_image *pixels;
    const int *luma;
    const static int luma1[16] = {
        2, 30, 10, 20, 12, 16, 8, 24, 12, 8, 16, 10, 15, 24, 15, 20
    };
    const static int luma2[16] = {
        2, 30, 8, 24, 16, 16, 8, 24, 16, 8, 16, 8, 16, 24, 16, 24
    };
    const static signed char cols[16][16] = {
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 0
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 1
        { -25, -5, 15, 34, 47, 53, 51, 42, 25, 5, -15, -34, -47, -53, -51, -42 }, // 2
        { 25, 7, -12, -30, -43, -49, -48, -40, -25, -7, 12, 30, 43, 49, 48, 40 }, // 3
        { 30, 48, 58, 59, 52, 36, 14, -8, -30, -48, -58, -59, -52, -36, -14, 8 }, // 4
        { -17, -36, -50, -56, -53, -43, -26, -4, 17, 36, 50, 56, 53, 43, 26, 4 }, // 5
        { 60, 55, 42, 23, 0, -23, -42, -55, -60, -55, -42, -23, 0, 23, 42, 55 }, // 6
        { -61, -57, -43, -23, 0, 23, 43, 57, 61, 57, 43, 23, 0, -23, -43, -57 }, // 7
        { -41, -21, 1, 23, 42, 55, 59, 54, 41, 21, -1, -23, -42, -55, -59, -54 }, // 8
        { -56, -41, -20, 4, 27, 47, 59, 62, 56, 41, 20, -4, -27, -47, -59, -62 }, // 9
        { -25, -5, 15, 34, 47, 53, 51, 42, 25, 5, -15, -34, -47, -53, -51, -42 }, // 10
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 11
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }, // 12
        { -17, -36, -50, -56, -53, -43, -26, -4, 17, 36, 50, 56, 53, 43, 26, 4 }, // 13
        { 60, 55, 42, 23, 0, -23, -42, -55, -60, -55, -42, -23, 0, 23, 42, 55 }, // 14
        { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }   // 15
    };
    unsigned char *p, *d;
    int x, y, w, lb, n2;
    signed char *l;
    unsigned int *offs = NULL;
    d = rscreen->d[frame];
    pixels = new_tv_pixels(width_tvpicture(rscreen->w, vmode, scale), rscreen->h, (((rscreen->w + 4) * scale * 9 / 4 + 2) / 3) * 3 + 6, scale);
    w = rscreen->w + 4;
    lb = 2;
    n2 = ((((pixels->pitch / 3) - pixels->w) / 2) * 3000 - lb * scale * 2250) + (scale == 1 ? 2000 : 3000);
    n2 += ((((pixels->pitch / 3) - pixels->w) / 2 + pixels->w) * 3000 - (lb + rscreen->w) * scale * 2250) + (scale == 1 ? 2000 : 3000);
    n2 /= 2;

    l = (signed char *)malloc((pixels->pitch + 3 * scale) * 196 / 63 / scale * 7 + 13 + 16);
    if (l == NULL) return NULL;
    memset(l, 0, (pixels->pitch + 3 * scale) * 196 / 63 / scale * 7 + 13 + 16);

    if (!(vmode & VIDEO_MONOCHROME)) {
        int a, i2, j;
        int scale2 = scale * 8;
        offs = (unsigned int *)malloc(sizeof(unsigned int *) * ((pixels->pitch / 3) + 2 * scale));
        if (offs == NULL) {
            free(l);
            return NULL;
        }
        for (j = i2 = a = x = 0; x < (pixels->pitch / 3) + 2 * scale; x++) {
            offs[x] = j;
            a += (196 * 3) % 63;
            if (a >= 63) {
                a -= 63;
                i2 += (196 * 3) / 63 + 1;
            } else i2 += (196 * 3) / 63;
            while (i2 >= scale2) {
                i2 -= scale2;
                j += 8;
            }
        }
    }

    luma = (vmode & VIDEO_OLD_LUMA) ? luma2 : luma1;
    memset(pixels->d, 0, pixels->pitch * pixels->h);
    for (y = 0; y < rscreen->h; y++) {
        int i, n = n2, ci = 0;
        int accu = 0;
        signed char *l2 = l;
        p = pixels->d + y * pixels->pitch;
        {
            int l;
            const signed char *c;
            int cb = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
            c = cols[cb];
            for (l = 0; l < 13; l++) {
                l2[ci] = c[ci];
                ci++;
            }
        }
        for (x = -lb; x < w - lb; x++) {
            int l, ll;
            const signed char *c;
            if (x < 0) {
                int cb = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
                if (rscreen->flags & VIEW64_SCREEN_BORDERLESS) cb = *d;
                l = luma[cb];
                c = cols[cb];
            } else if (x >= rscreen->w) {
                int cb = rscreen->b ? rscreen->b[y * 2 + 1] : rscreen->br;
                if (rscreen->flags & VIEW64_SCREEN_BORDERLESS) cb = d[-1];
                l = luma[cb];
                c = cols[cb];
            } else {
                l = luma[*d];
                c = cols[*d];
                d++;
            }
            accu += (1000 - n) * l;
            *p++ = accu / 128;
            n += 2250 * scale - 1000;
            ll = 1000 * l / 128;
            while (n >= 1000) {
                *p++ = ll;
                n -= 1000;
            }
            accu = n * l;
            if (c[0]) {
                for (l = 0; l < 7; l++) {
                    l2[ci] = c[ci & 15];
                    ci++;
                }
            } else {
                memset(l2 + ci, 0, 7);
                ci += 7;
            }
        }
        p = pixels->d + y * pixels->pitch;
        if (scale == 1) {
            int a = 0, b;
            a = b = p[2];
            for (x = 0; x < 4; x++) a += p[x];
            for (x = 0; x < pixels->pitch - 4; x++) {
                a += p[x+4] - b;
                b = p[x];
                p[x] = a / 5;
            }
        } else {
            int a = 0, b;
            a = b = p[2];
            for (x = 0; x < 6; x++) a += p[x];
            for (x = 0; x < pixels->pitch - 6; x++) {
                a += p[x+6] - b;
                b = p[x];
                p[x] = a / 7;
            }
        }
        if (!(vmode & VIDEO_MONOCHROME)) {
            unsigned int *offs1 = offs, *offs2 = offs;
            int r = 0, g = 0, b = 0;
            for (x = 0; x < 2 * scale; x++) {
                int j = *offs2++;
                int e = l[j + 4];
                if (j & 8) {
                    r -= e;
                    g += e;
                    b += l[j + 8];
                } else {
                    r += e;
                    g -= e;
                    b -= l[j + 8];
                }
            }
            for (x = 0; x < pixels->pitch; x += 3) {
                int j;
                int v;
                int j2 = *offs1++;
                int e = l[j2 + 4];
                if (j2 & 8) {
                    r += e;
                    g -= e;
                    b -= l[j2 + 8];
                } else {
                    r -= e;
                    g += e;
                    b += l[j2 + 8];
                }
                j = *offs2++;
                e = l[j + 4];
                if (j & 8) {
                    r -= e;
                    g += e;
                    b += l[j + 8];
                } else {
                    r += e;
                    g -= e;
                    b -= l[j + 8];
                }
                v = p[x];
                v += r / 2 / scale;
                p[x] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                v = p[x+1];
                v += g * 2 / 5 / scale;
                p[x+1] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
                v = p[x+2];
                v += b * 2 / 3 / scale;
                p[x+2] = v & ~255 ? ((v < 0) ? 0 : 255) : v;
            }
        }
        if (vmode & VIDEO_CHROMA_NOISE) {
            int a, j;

            for (j = a = i = x = 0; x < pixels->pitch; x++) {
                int v = p[x] << 4;
                v += l[j + 4] + 8;
                p[x] = (v <= 0) ? 0 : (v >= (255 << 4)) ? 255 : (v >> 4);
                a += 196 % 63;
                if (a >= 63) {
                    a -= 63;
                    i += (196 / 63) + 1;
                } else i += 196 / 63;
                while (i >= scale) {
                    i -= scale; j++;
                }
            }
        }
    }
    free(offs);
    free(l);
    return pixels;
}

static tvrender_image *create_rgb(const view64_screen *rscreen, int scale, int frame, const rgb_color *palette, unsigned int vmode) {
    tvrender_image *pixels;
    unsigned char *p, *d, *e;
    int x, y;
    d = rscreen->d[frame];
    pixels = new_tv_pixels(width_tvpicture(rscreen->w, vmode, scale), rscreen->h, (rscreen->w + 2) * scale * 3, scale);
    for (y = 0; y < rscreen->h; y++) {
        const rgb_color *rgb2 = &palette[rscreen->b ? rscreen->b[y * 2] : rscreen->br];
        p = pixels->d + y * pixels->pitch;
        if (vmode & VIDEO_MONOCHROME) {
            memset(p, (rgb2->r * 114 + rgb2->g * 142) / 256, scale * 3);
            p += 3 * scale;
            for (x = 0; x < rscreen->w; x++) {
                rgb2 = &palette[d[x]];
                memset(p, (rgb2->r * 114 + rgb2->g * 142) / 256, scale * 3);
                p += scale * 3;
            }
            d += x;
            rgb2 = &palette[rscreen->b ? rscreen->b[y * 2 + 1] : rscreen->br];
            memset(p, (rgb2->r * 114 + rgb2->g * 142) / 256, scale * 3);
        } else {
            unsigned char rgb[3];
            rgb[0] = rgb2->r; rgb[1] = rgb2->g; rgb[2] = rgb2->b;
            for (e = p + scale * 3; p < e; p += 3) {
                memcpy(p, rgb, 3);
            }
            for (x = 0; x < rscreen->w; x++) {
                rgb2 = &palette[d[x]];
                rgb[0] = rgb2->r; rgb[1] = rgb2->g; rgb[2] = rgb2->b;
                for (e += scale * 3; p < e; p += 3) {
                    memcpy(p, rgb, 3);
                }
            }
            d += x;
            rgb2 = &palette[rscreen->b ? rscreen->b[y * 2 + 1] : rscreen->br];
            rgb[0] = rgb2->r; rgb[1] = rgb2->g; rgb[2] = rgb2->b;
            for (e = p + scale * 3; p < e; p += 3) {
                memcpy(p, rgb, 3);
            }
        }
    }
    return pixels;
}

static inline unsigned char intensity_mix(unsigned char a, unsigned char b) {
    return sqrt((a * a + b * b) / 2);
}

static tvrender_image *create_tvpicture_truecolor(const view64_screen *rscreen, unsigned int vmode, int scale, const rgb_color *palette) {
    int x, y;
    unsigned char *p, *p2;
    tvrender_image *pixels, *pixels2;

    switch (vmode & VIDEO_MASK) {
    case VIDEO_NTSC: pixels = create_ntsc(rscreen, scale, 0, vmode); break;
    case VIDEO_PAL: pixels = create_pal(rscreen, scale, 0, vmode); break;
    default: pixels = create_rgb(rscreen, scale, 0, palette, vmode); break;
    }
    if (pixels == NULL) return NULL;
    if (rscreen->n == 2) { /* interlace */
        int i;
        switch (vmode & VIDEO_MASK) {
        case VIDEO_NTSC: pixels2 = create_ntsc(rscreen, scale, 1, vmode); break;
        case VIDEO_PAL: pixels2 = create_pal(rscreen, scale, 1, vmode); break;
        default: pixels2 = create_rgb(rscreen, scale, 1, palette, vmode); break;
        }
        if (pixels2 == NULL) {
            free_tv_pixels(pixels);
            return NULL;
        }
        p = pixels->d;
        p2 = pixels2->d;
        if ((vmode & VIDEO_NOLACE) && (vmode & VIDEO_MASK) == VIDEO_RGB && (rscreen->flags & VIEW64_SCREEN_MCI)) {
            for (y = 0; y < pixels->h; y++) {
                for (x = 0; x < pixels->pitch; x += 6 * scale) {
                    memcpy(p + x, p2 + x, 3 * scale);
                }
                p += pixels->pitch;
                p2 += pixels2->pitch;
            }
        } else {
            static unsigned short indexes[256];
            static unsigned char cache[32896];
            if (indexes[1] == 0) {
                for (x = i = 0; i < 256; i++, x += i) indexes[i] = x;
            }
            for (i = 0; i < pixels->pitch * pixels->h; i++) {
                int k = p[i] >= p2[i] ? indexes[p[i]] + p2[i] : indexes[p2[i]] + p[i];
                if (cache[k] == 0 && k > 1) {
                    cache[k] = intensity_mix(p[i], p2[i]);
                }
                p[i] = cache[k];
            }
        }
        free_tv_pixels(pixels2);
    }
    return pixels;
}

int colors_tvpicture(const view64_screen *rscreen, unsigned int vmode) {
    if ((vmode & (VIDEO_SCANLINES | VIDEO_SHADOWMASK)) || (vmode & VIDEO_MASK) != VIDEO_RGB) {
        return 256*256*256;
    }
    if (rscreen->n == 2) { /* interlace */
        if ((vmode & VIDEO_NOLACE) && (vmode & VIDEO_MASK) == VIDEO_RGB && (rscreen->flags & VIEW64_SCREEN_MCI)) {
            return 16;
        }
        return 256;
    }
    return 16;
}

int width_tvpicture(int w, unsigned int vmode, int scale) {
    switch (vmode & VIDEO_MASK) {
    case VIDEO_NTSC: return w * scale * 3 / 4;
    case VIDEO_PAL: return w * scale * 187 / 200;
    default: return w * scale;
    }
}

tvrender_image *create_tvpicture(const view64_screen *rscreen, unsigned int vmode, int scale, const rgb_color *palette) {
    int x, y;
    unsigned char *p, *r;
    tvrender_image *pixels;

    if (palette == NULL) {
        palette = (vmode & VIDEO_OLD_LUMA) ? pal_old_palette : pal_palette;
    }

    if ((vmode & (VIDEO_SCANLINES | VIDEO_SHADOWMASK)) || (vmode & VIDEO_MASK) != VIDEO_RGB) {
        return create_tvpicture_truecolor(rscreen, vmode, scale, palette);
    }

    pixels = new_tv_pixels(rscreen->w, rscreen->h, rscreen->w + 2, scale);
    if (pixels == NULL) return NULL;
    p = pixels->d;
    r = rscreen->d[0];
    if (rscreen->n == 2) { /* interlace */
        unsigned char *r2 = rscreen->d[1];
        if ((vmode & VIDEO_NOLACE) && (vmode & VIDEO_MASK) == VIDEO_RGB && (rscreen->flags & VIEW64_SCREEN_MCI)) {
            for (y = 0; y < pixels->h; y++) {
                x = 0;
                p[x] = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
                while (x < rscreen->w) {
                    unsigned char e = r[x];
                    x++;
                    p[x] = e;
                    if (x >= rscreen->w) break;
                    e = r2[x];
                    x++;
                    p[x] = e;
                }
                x++;
                p[x] = rscreen->b ? rscreen->b[y * 2 + 1] : rscreen->br;
                p += pixels->pitch;
                r += rscreen->w;
                r2 += rscreen->w;
            }
        } else {
            if (pixels->palette == NULL) {
                rgb_color *rgb;
                pixels->colors = 256;
                pixels->palette = (rgb_color *)malloc(pixels->colors * sizeof(rgb_color));
                if (pixels->palette == NULL) {
                    free_tv_pixels(pixels);
                    return NULL;
                }
                rgb = pixels->palette;
                for (y = 0; y < 16; y++) {
                    for (x = 0; x < 16; x++) {
                        if (x < y) {
                            *rgb = pixels->palette[y + x * 16];
                        } else if (x == y) {
                            if (vmode & VIDEO_MONOCHROME) {
                                rgb->r = rgb->g = rgb->b = (palette[x].r * 114 + palette[x].g * 142) / 256;
                            } else {
                                *rgb = palette[x];
                            }
                        } else {
                            if (vmode & VIDEO_MONOCHROME) {
                                rgb->r = rgb->g = rgb->b = (intensity_mix(palette[x].r, palette[y].r) * 114 + intensity_mix(palette[x].g, palette[y].g) * 142) / 256;
                            } else {
                                rgb->r = intensity_mix(palette[x].r, palette[y].r);
                                rgb->g = intensity_mix(palette[x].g, palette[y].g);
                                rgb->b = intensity_mix(palette[x].b, palette[y].b);
                            }
                        }
                        rgb++;
                    }
                }
            }
            for (y = 0; y < pixels->h; y++) {
                x = 0;
                p[x] = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
                while (x < rscreen->w) {
                    unsigned char e = r[x] | (r2[x] << 4);
                    x++;
                    p[x] = e;
                }
                x++;
                p[x] = rscreen->b ? rscreen->b[y * 2 + 1] : rscreen->br;
                p += pixels->pitch;
                r += rscreen->w;
                r2 += rscreen->w;
            }
        }
    } else {
        for (y = 0; y < pixels->h; y++) {
            p[0] = rscreen->b ? rscreen->b[y * 2] : rscreen->br;
            memcpy(p + 1, r, rscreen->w);
            p[rscreen->w + 1] = rscreen->b ? rscreen->b[y * 2 + 1] : rscreen->br;
            p += pixels->pitch;
            r += rscreen->w;
        }
    }
    if (pixels->palette == NULL) {
        pixels->colors = 16;
        pixels->palette = (rgb_color *)malloc(pixels->colors * sizeof(rgb_color));
        if (pixels->palette == NULL) {
            free_tv_pixels(pixels);
            return NULL;
        }
        if (vmode & VIDEO_MONOCHROME) {
            for (x = 0; x < 16; x++) {
                rgb_color *rgb = &pixels->palette[x];
                rgb->r = rgb->g = rgb->b = (palette[x].r * 114 + palette[x].g  * 142) / 256;
            }
        } else {
            memcpy(pixels->palette, palette, pixels->colors * sizeof(rgb_color));
        }
    }
    return pixels;
}

int create_tvpicture_line(const tvrender_image *pixels, unsigned int vmode, int x, int y, int w, unsigned char *line) {
    unsigned char *p, *p2;
    int scale = pixels->scale;

    if (pixels->palette) {
        unsigned char *e, *e2;
        if (w == 0) return 0;
        p2 = pixels->d + ((pixels->pitch - pixels->w) / 2 + x / scale) + pixels->pitch * (y / scale);
        e2 = line + (scale - (x % scale)) * 3; e = line + w * 3;
        for (;;) {
            unsigned char rgb[3];
            rgb_color *rgb_palette = &pixels->palette[*p2];
            rgb[0] = rgb_palette->r; rgb[1] = rgb_palette->g; rgb[2] = rgb_palette->b;
            if (e2 < e) {
                while (line < e2) {
                    memcpy(line, rgb, 3);
                    line += 3;
                }
            } else {
                while (line < e) {
                    memcpy(line, rgb, 3);
                    line += 3;
                }
                return 0;
            }
            e2 += scale * 3;
            p2++;
        }
    }

    if ((vmode & VIDEO_SCANLINES) && scale > 1 && scale <= 13) { /* render scanlines */
        int a[13*3], *t;
        static int t1[13 * 13];
        int r = 0;
        static int scale_cache;
        static unsigned char loop[13 * 3];
        if (scale_cache != scale) {
            int x1, y1, *t2;
            for (y1 = 0; y1 < scale * 3; y1++) {
                loop[y1] = ((y1 / 3 + 1) % scale) * 3 + (y1 % 3);
            }
            t2 = t1;
            for (y1 = 0; y1 < scale; y1++) {
                int il = (y1 * 2 - scale) * 2 + 3;
                il = scale * scale * 4 - il * il;
                for (x1 = 0; x1 < scale; x1++) {
                    int il2 = (x1 * 2 - scale) * 2 + 3;
                    il2 = scale * scale * 4 - il2 * il2 + il;
                    il2 += scale * scale * 4;
                    if (il2 < 0) il2 = 0;
                    else il2 = il2 * 26 / scale / scale / scale;
                    *t2++ = il2;
                }
            }
            scale_cache = scale;
        }
        t = &t1[(y % scale) * scale];

        p2 = pixels->d + ((pixels->pitch / 3 - pixels->w) / 2 + x) * 3 + pixels->pitch * (y / scale) + scale / 2 * 3;
        p = line;
        memset(a, 0, scale * sizeof(a[0]) * 3);
        for (x = -scale * 3; x < 0; x += 3) {
            int r2 = r;
            int i1 = p2[x], i2 = p2[x + 1], i3 = p2[x + 2];
            int x1;
            for (x1 = 0; x1 < scale; x1++) {
                int i = t[x1];
                a[r2] += i1 * i;
                a[r2 + 1] += i2 * i;
                a[r2 + 2] += i3 * i;
                r2 = loop[r2];
            }
            a[r] = a[r + 1] = a[r + 2] = 0;
            r = loop[r];
        }
        for (; x < w * 3; x += 3) {
            int r2 = r;
            int i1 = p2[x], i2 = p2[x + 1], i3 = p2[x + 2];
            int x1;
            for (x1 = 0; x1 < scale; x1++) {
                int i = t[x1];
                a[r2] += i1 * i;
                a[r2 + 1] += i2 * i;
                a[r2 + 2] += i3 * i;
                r2 = loop[r2];
            }
            i1 = a[r];
            p[x] = (i1 < 65536) ? (i1 >> 8) : 255; a[r] = 0;
            i1 = a[r + 1];
            p[x + 1] = (i1 < 65536) ? (i1 >> 8) : 255; a[r + 1] = 0;
            i1 = a[r + 2];
            p[x + 2] = (i1 < 65536) ? (i1 >> 8) : 255; a[r + 2] = 0;
            r = loop[r];
        }
    } else { /* no scanlines */
        p2 = pixels->d + ((pixels->pitch / 3 - pixels->w) / 2 + x) * 3 + pixels->pitch * (y / scale);
        memcpy(line, p2, w * 3);
    }
    if (vmode & VIDEO_SHADOWMASK) { /* render shadow mask */
        int x2 = x / 3, n = x % 3;
        static unsigned char c78[256];
        static unsigned char c98[256];
        if (c78[255] == 0) {
            for (y = 0; y < 256; y++) {
                c78[y] = (y * 7) >> 3;
                c98[y] = ((y * 9) >> 3) > 255 ? 255 : ((y * 9) >> 3);
            }
        }

        p = line;
        for (x = 0; x < w; x++) {
            int i = (y & 3) == (x2 & 1) * 2;
            switch (n) {
            case 0:
                p[0] = i ? c78[p[0]] : c98[p[0]];
                p[1] = c78[p[1]];
                p[2] = c78[p[2]];
                n = 1;
                break;
            case 1:
                p[0] = c78[p[0]];
                p[1] = i ? c78[p[1]] : c98[p[1]];
                p[2] = c78[p[2]];
                n = 2;
                break;
            case 2:
                p[0] = c78[p[0]];
                p[1] = c78[p[1]];
                p[2] = i ? c78[p[2]] : c98[p[2]];
                n = 0; x2++;
                break;
            }
            p += 3;
        }
    }
    return 0;
}

rgb_color *load_palette(FILE *f) {
    unsigned char values[4];
    int line = 0;
    int i, j, e = 0, c = '\n';
    rgb_color *palette;

    if (f == NULL) {
        return NULL;
    }
    palette = (rgb_color *)calloc(16, sizeof(rgb_color));
    if (palette == NULL) {
        fprintf(stderr, "Not enough memory for palette\n");
        return NULL;
    }

    while (c != EOF && e < 16) {
        line++;
        memset(values, 0, sizeof(values));
        for (i = 0; i < 4; i++) {
            int v = 0;

            do {
                c = fgetc(f);
            } while (c == ' ' || c == '\t');

            if (c == '#' || c == '\n' || c == EOF) {
                while (c != '\n' && c != EOF) {
                    c = fgetc(f);
                }
                break;
            }

            while (c == '0') {
                c = fgetc(f);
            }

            for (j = 0; j < (2 + (i < 3)); j++) {
                if (c >= '0' && c <= '9') {
                    v = v * 16 + c - '0';
                } else if ((c | 0x20) >= 'a' && (c | 0x20) <= 'f') {
                    v = v * 16 + (c | 0x20) - 'a' + 10;
                } else {
                    break;
                }
                c = fgetc(f);
            }
            values[i] = v;
            if (j == 2 + (i < 3)) {
                fprintf(stderr, "%d: Value out of range\n", line);
            }
            while (c != ' ' && c != '\t' && c != '\n' && c != EOF) {
                c = fgetc(f);
            }
        }
        if (i > 0 && i < 4) {
            fprintf(stderr, "%d: Not enough entries\n", line);
        }
        if (c != '#' && c != '\n' && c != EOF) {
            fprintf(stderr, "%d: Junk at end of line\n", line);
        }
        if (i == 4) {
            palette[e].r = values[0];
            palette[e].g = values[1];
            palette[e].b = values[2];
            e++;
        }
    }
    if (e < 16) {
        fprintf(stderr, "Not enough colors\n");
    }
    return palette;
}
