/*
    View64 configuration handler
    (c) 1995-2013 Soci/Singular (soci at c64.rulez.org)

    $Id: view64_config.c 96 2017-08-22 19:37:49Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include <string.h>
#include "getopt.h"
#include "view64_config.h"
#include "main.h"

static const char *short_options= "cfh:ilmup:stw:x:y:z:NPRV?";

static const struct option long_options[]={
    {"chroma-noise"     , no_argument      , 0, 'c'},
    {"full-screen"      , no_argument      , 0, 'f'},
    {"no-menu-bar"      , no_argument      , 0,  6},
    {"height"           , required_argument, 0, 'h'},
    {"mci-deinterlace"  , no_argument      , 0, 'i'},
    {"monochrome"       , no_argument      , 0, 'm'},
    {"old-luma"         , no_argument      , 0, 'u'},
    {"ntsc"             , no_argument      , 0, 'N'},
    {"pal"              , no_argument      , 0, 'P'},
    {"rgb"              , no_argument      , 0, 'R'},
    {"palette"          , required_argument, 0, 'p'},
    {"save"             , no_argument      , 0, 's'},
    {"scanlines"        , no_argument      , 0, 'l'},
    {"tv-mask"          , no_argument      , 0, 't'},
    {"width"            , required_argument, 0, 'w'},
    {"version"          , no_argument      , 0, 'V'},
    {"usage"            , no_argument      , 0,   2},
    {"win-x"            , required_argument, 0, 'x'},
    {"win-y"            , required_argument, 0, 'y'},
    {"zoom"             , required_argument, 0, 'z'},
    {"help"             , no_argument      , 0,   3},
    {"help-formats"     , no_argument      , 0,   4},
    {"help-keys"        , no_argument      , 0,   5},
    { 0, 0, 0, 0}
};

static int testarg(int argc, char *argv[], arguments_t *arguments) {
    int opt, longind;

    while ((opt = getopt_long(argc, argv, short_options, long_options, &longind)) != -1)
        switch (opt)
        {
            case 'c': arguments->vmode |= VIDEO_CHROMA_NOISE; break;
            case 'f': arguments->fullscreen = 1; break;
            case 'h': arguments->height = atoi(optarg); if (arguments->height < 0) arguments->height = 0; break;
            case 'i': arguments->vmode |= VIDEO_NOLACE; break;
            case 'l': arguments->vmode |= VIDEO_SCANLINES; break;
            case 'm': arguments->vmode |= VIDEO_MONOCHROME; break;
            case 'u': arguments->vmode |= VIDEO_OLD_LUMA; break;
            case 'p':
                free(arguments->palette_filename);
                arguments->palette_filename = (char *)malloc(strlen(optarg) + 1);
                if (arguments->palette_filename) {
                    strcpy(arguments->palette_filename, optarg);
                }
                break;
            case 6: arguments->menu = 0; break;
            case 's': arguments->save = 1; break;
            case 't': arguments->vmode |= VIDEO_SHADOWMASK; break;
            case 'w': arguments->width = atoi(optarg); if (arguments->width < 0) arguments->width = 0;break;
            case 'x': arguments->x = atoi(optarg); if (arguments->x < -1) arguments->x = -1; break;
            case 'y': arguments->y = atoi(optarg); if (arguments->y < -1) arguments->y = -1; break;
            case 'z': arguments->zoom = atoi(optarg); if (arguments->zoom < 0) arguments->zoom = 0;break;
            case 'N': arguments->vmode = (arguments->vmode & ~VIDEO_MASK) | VIDEO_NTSC; break;
            case 'P': arguments->vmode = (arguments->vmode & ~VIDEO_MASK) | VIDEO_PAL; break;
            case 'R': arguments->vmode = (arguments->vmode & ~VIDEO_MASK) | VIDEO_RGB; break;
            case 2:puts(
                 /* 12345678901234567890123456789012345678901234567890123456789012345678901234567890 */
                   "Usage: view64 [-cfilmstuNPRV?] [-h <height>] [-w <width>] [-x <pixel>]\n"
                   "       [-y <pixel>] [-z <zoom>] [-p <palette>] [--width <pixels>]\n"
                   "       [--height <pixels>] [--win-x <pixels>] [--win-y <pixels>]\n"
                   "       [--zoom <level>] [--full-screen] [--chroma-noise] [--mci-deinterlace]\n"
                   "       [--ntsc] [--pal] [--rgb] [--palette <file>] [--scanlines] [--tv-mask]\n"
                   "       [--monochrome] [--old-luma] [--save] [--help] [--help-formats]\n"
                   "       [--help-keys] [--usage] [--version] PICTURE\n"
                   );
                   return 0;

            case 'V': puts("View64 " VERSION " (C) 1995-" YEAR " Soci/Singular (Kajtar Zsolt)\n");
                     return 0;
            case 4: puts("View64 " VERSION " supports the following formats:\n"
                         "\n"
                         "*.hpi - Art Studio v1.1\n"
                         "*.mpi - Art Studio v1.2b\n"
                         "*.mwi - Art Studio v1.2b window\n"
                         "*.drl - Draz-lace v1.0\t\t\t\tPacked/Unpacked\n"
                         "*.ifl - Gunpaint IFLI\n"
                         "*.afl - AFLI-editor v2.1\n"
                         "*.him - Hires manager v1.0\t\t\tPacked/Unpacked\n"
                         "*.eci - ECIgraphic-editor v1.0\t\t\tPacked/Unpacked\n"
                         "*.fli - Fli designer v2.0\n"
                         "*.flp - Fli Profi v1.0\n"
                         "*.flg - Fli Graph v2.2\t\t\t\tPacked/Unpacked\n"
                         "*.fbi - Flip v1.33\n"
                         "*.fun - Funpaint II\t\t\t\tPacked/Unpacked\n"
                         "*.pp  - Pixel perfect\t\t\t\tPacked/Unpacked\n"
                         "*.hcb - HCB-editor v0.05\n"
                         "*.drp - Drazpaint v1.3/1.5/2.0\t\t\tPacked/Unpacked\n"
                         "*.kla - Koala painter\t\t\t\tPacked/Unpacked\n"
                         "*.fcp - Face-painter v1.0\n"
                         "*.ami - Amica paint 1.0\n"
                         "*.bpl - Blazing Paddles\n"
                         "*.che - Cheese paint V2\n"
                         "*.cwg - Create with Garfield\n"
                         "*.dol - Dolphin Ed\n"
                         "*.mil - Micro illustrator\t\t\tPacked/Unpacked\n"
                         "*.shf - Super Hires Fli editor v1.0\t\tPacked/Unpacked\n"
                         "*.shi - Super Hires Interlace editor v1.3\tPacked/Unpacked\n"
                         "*.esh - Extend Super Hires Interlace editor\tPacked/Unpacked\n"
                         "*.ufl - UFLI-editor v1.0/v2.0\t\t\tPacked/Unpacked\n"
                         "*.uif - UIFLI-editor v1.0\t\t\tPacked/Unpacked\n"
                         "*.shx - SHF-XL edit v1.0\t\t\tPacked/Unpacked\n"
                         "*.xfl - X-FLI editor v2.1\t\t\tPacked/Unpacked\n"
                         "*.muf - MUFLI editor v1.0\t\t\tPacked/Unpacked\n"
                         "*.mui - MUIFLI editor v0.2\n"
                         "*.nuf - NUFLI editor v1.0\t\t\tPacked/Unpacked\n"
                         "*.sif - Super Hires Interlace Fli editor v1.0\n"
                         "*.ish - Image System hires\n"
                         "*.ism - Image System multi\n"
                         "*.mci - Truepaint I\t\t\t\tPacked/Unpacked\n"
                         "*.cdu - CDU Paint\n"
                         "*.hfd - Hires FLI Design\n"
                         "*.ddp - Doodle paint\t\t\t\tPacked/Unpacked\n"
                         "*.vid - Vidcom 64\n"
                         "*.lre - Interpaint 1.0 multi\n"
                         "*.hre - Interpaint 1.0 hires\n"
                         "*.a64 - Artist 64\n"
                         "*.sar - Saracen paint\n"
                         "*.hed - Hi-eddi\n"
                         "*.pmg - Paint magic, Paint magic+\n"
                         "*.zom - Zoomatic\n"
                         "*.hie - Hires interlace editor v1.0\n"
                         "*.het - Hires-editor v1.0\n"
                         "*.fed - FLI editor/Saga\n"
                         "*.flm - Flimatic v3.7\t\t\t\tPacked/Unpacked\n"
                         "*.sh1 - Super-hires editor I\t\t\tPacked/Unpacked\n"
                         "*.sh2 - Super-hires editor II\t\t\tPacked/Unpacked\n"
                         "*.she - Super hires editor\n"
                         "*.she - Super hires editor 2\n"
                         "*.gig - Giga paint 1.2 hires/multi\n"
                         "*.emc - EMC-editor\n"
                         "*.hle - Hireslace editor v1.5\n"
                         "*.shs - Super hires studio\n"
                         "*.wig - Wigmore artist\n"
                         "*.4bt - 4 bit GoDot\t\t\t\tPacked/Unpacked\n"
                         "*.clp - 4 bit GoDot clip\t\t\tPacked/Unpacked\n"
                         "*.vbm - VBM bitmap v2\n"
                         "*.rp  - Rainbow painter\n"
                         "*.rpm - Run paint multi\n"
                         "*.rph - Run paint hires\n"
                         "*.rpo - Run paint mono\n"
                         "*.ish - Interlace-Super-Hires Painter v1.0\n"
                         "*.p64 - Picasso 64\n"
                         "*.fly - Flying colors\n"
                         "*.lp3 - Logo-Painter V3/V3+\n"
                         "*.cle - Centauri logo-editor V1.2/V1.3\n"
                         "*.ile - Interlaced logo editor V1.01\n"
                         "\n"
                         "*.ffl - Flash FLI\n"
                         "*.bfl - Big FLI\n"
                         "*.ctm - Charpad CTM v1-v5\n"
                         "*.spd - Spritepad SPD\n");
                    return 0;
            case 5: puts("View64 " VERSION " window keys:\n"
                         "\n"
                         "escape - Quit\n"
                         "1-9    - Zoom level\n"
                         "c      - Toggle chroma noise rendering\n"
                         "f      - Toggle full-screen\n"
                         "i      - Toggle MCI deinterlacing\n"
                         "l      - Toggle scanlines rendering\n"
                         "m      - Toggle monochrome rendering\n"
                         "n      - Use NTSC decoder\n"
                         "p      - Use PAL decoder\n"
                         "r      - Use RGB decoder\n"
                         "s      - Save to BMP\n"
                         "t      - Toggle TV mask\n"
                         "u      - Toggle old luminance\n");
                    return 0;
            case 3:
            case '?':if (optopt=='?' || opt==3) { puts(
               "Usage: view64 [OPTIONS...] PICTURE\n"
               "View64 " VERSION " (C) 1995-" YEAR " Soci/Singular (Kajtar Zsolt)\n"
               "\n"
               " Window control:\n"
               "  -f, --full-screen     Start in full screen\n"
               "      --no-menu-bar     Start without menu\n"
               "  -h, --height <pixels> Window height\n"
               "  -w, --width <pixels>  Window width\n"
               "  -x, --win-x <pixels>  Window x position\n"
               "  -y, --win-y <pixels>  Window y position\n"
               "  -z, --zoom <level>    Zoom level\n"
               "\n"
               " Rendering options:\n"
               "  -c, --chroma-noise    Add chroma noise for PAL/NTSC\n"
               "  -i, --mci-deinterlace Fake multicolor deinterlacing (RGB only)\n"
               "  -p, --palette <file>  Palette file for RGB mode\n"
               "  -l, --scanlines       Render scanlines\n"
               "  -m, --monochrome      Monochrome\n"
               "  -t, --tv-mask         Render shadow mask\n"
               "  -u, --old-luma        Old luminance\n"
               "  -N, --ntsc            Use NTSC decoder\n"
               "  -P, --pal             Use PAL decoder\n"
               "  -R, --rgb             Use RGB decoder\n"
               "\n"
               " Misc:\n"
               "  -s, --save            Save BMP image\n"
               "  -?, --help            Give this help list\n"
               "      --help-formats    List known image formats\n"
               "      --help-keys       List window keys\n"
               "      --usage           Give a short usage message\n"
               "  -V, --version         Print program version\n"
               "\n"
               "Mandatory or optional arguments to long options are also mandatory or optional\n"
               "for any corresponding short options.\n"
               "\n"
               "View64 comes with ABSOLUTELY NO WARRANTY; This is free software, and you\n"
               "are welcome to redistribute it under certain conditions; See LICENSE!\n"
               "\n"
               "Report bugs to <soci" "\x40" "c64.rulez.org>.");
               return 0;
            }
            default:puts("Try `view64 --help' or `view64 --usage' for more information.\n");
                    return -1;
        }
    return optind;
}

int ini_handler(void *user, const char *section, const char *name, const char *value)
{
    arguments_t *arguments = (arguments_t *)user;
    if (strcmp(section, "view64")) {
        return 0;
    }
    if (!strcmp(name, "chroma-noise")) {
        arguments->vmode = atoi(value) ? (arguments->vmode | VIDEO_CHROMA_NOISE) : (arguments->vmode & ~VIDEO_CHROMA_NOISE);
    } else if (!strcmp(name, "mci-deinterlace")) {
        arguments->vmode = atoi(value) ? (arguments->vmode | VIDEO_NOLACE) : (arguments->vmode & ~VIDEO_NOLACE);
    } else if (!strcmp(name, "scanlines")) {
        arguments->vmode = atoi(value) ? (arguments->vmode | VIDEO_SCANLINES) : (arguments->vmode & ~VIDEO_SCANLINES);
    } else if (!strcmp(name, "tv-mask")) {
        arguments->vmode = atoi(value) ? (arguments->vmode | VIDEO_SHADOWMASK) : (arguments->vmode & ~VIDEO_SHADOWMASK);
    } else if (!strcmp(name, "monochrome")) {
        arguments->vmode = atoi(value) ? (arguments->vmode | VIDEO_MONOCHROME) : (arguments->vmode & ~VIDEO_MONOCHROME);
    } else if (!strcmp(name, "old-luma")) {
        arguments->vmode = atoi(value) ? (arguments->vmode | VIDEO_OLD_LUMA) : (arguments->vmode & ~VIDEO_OLD_LUMA);
    } else if (!strcmp(name, "palette")) {
        free(arguments->palette_filename);
        arguments->palette_filename = (char *)malloc(strlen(value) + 1);
        if (arguments->palette_filename) {
            strcpy(arguments->palette_filename, value);
        }
    } else if (!strcmp(name, "decoder")) {
        if (!strcmp(value, "ntsc")) {
            arguments->vmode = (arguments->vmode & ~VIDEO_MASK) | VIDEO_NTSC;
        } else if (!strcmp(value, "pal")) {
            arguments->vmode = (arguments->vmode & ~VIDEO_MASK) | VIDEO_PAL;
        } else if (!strcmp(value, "rgb")) {
            arguments->vmode = (arguments->vmode & ~VIDEO_MASK) | VIDEO_RGB;
        }
    } else if (!strcmp(name, "x")) {
        arguments->x = atoi(value);  if (arguments->x < -1) arguments->x = -1;
    } else if (!strcmp(name, "y")) {
        arguments->y = atoi(value);  if (arguments->y < -1) arguments->y = -1;
    } else if (!strcmp(name, "width")) {
        arguments->width = atoi(value);  if (arguments->width < 0) arguments->width = 0;
    } else if (!strcmp(name, "height")) {
        arguments->height = atoi(value);  if (arguments->height < 0) arguments->height = 0;
    } else if (!strcmp(name, "zoom")) {
        arguments->zoom = atoi(value); if (arguments->zoom < 0) arguments->zoom = 0;
    } else if (!strcmp(name, "full-screen")) {
        arguments->fullscreen = !!atoi(value); 
    } else if (!strcmp(name, "menu-bar")) {
        arguments->menu = !!atoi(value); 
    } else {
        return 0;
    }
    return 1;
}

static arguments_t old_arguments;

int config_changed(const arguments_t *arguments) {
    if (arguments->vmode == old_arguments.vmode &&
            arguments->zoom == old_arguments.zoom &&
            arguments->fullscreen == old_arguments.fullscreen &&
            arguments->menu == old_arguments.menu &&
            arguments->x == old_arguments.x &&
            arguments->y == old_arguments.y &&
            arguments->width == old_arguments.width &&
            arguments->height == old_arguments.height) {

        if (arguments->palette_filename == old_arguments.palette_filename || 
                (arguments->palette_filename && old_arguments.palette_filename
                 && !strcmp(arguments->palette_filename,
                     old_arguments.palette_filename))) {
            return 0; /* no change */
        }
    }
    return 1;
}

void ini_save_file(FILE *f1, const arguments_t *arguments) {
    fputs("[view64]", f1);
    fprintf(f1, "\nx=%d", arguments->x);
    fprintf(f1, "\ny=%d", arguments->y);
    fprintf(f1, "\nwidth=%d", arguments->width);
    fprintf(f1, "\nheight=%d", arguments->height);
    fprintf(f1, "\nzoom=%d", arguments->zoom);
    fputs("\nfull-screen=", f1); putc(arguments->fullscreen ? '1' : '0', f1);
    fputs("\nmenu-bar=", f1); putc(arguments->menu ? '1' : '0', f1);
    fputs("\nchroma-noise=", f1); putc((arguments->vmode & VIDEO_CHROMA_NOISE) ? '1' : '0', f1);
    fputs("\nmci-deinterlace=", f1); putc((arguments->vmode & VIDEO_NOLACE) ? '1' : '0', f1);
    fputs("\nscanlines=", f1); putc((arguments->vmode & VIDEO_SCANLINES) ? '1' : '0', f1);
    fputs("\ntv-mask=", f1); putc((arguments->vmode & VIDEO_SHADOWMASK) ? '1' : '0', f1);
    fputs("\nmonochrome=", f1); putc((arguments->vmode & VIDEO_MONOCHROME) ? '1' : '0', f1);
    fputs("\nold-luma=", f1); putc((arguments->vmode & VIDEO_OLD_LUMA) ? '1' : '0', f1);
    fputs("\npalette=", f1); if (arguments->palette_filename) fputs(arguments->palette_filename, f1);
    fputs("\ndecoder=", f1);
    switch (arguments->vmode & VIDEO_MASK) {
    case VIDEO_NTSC: fputs("ntsc", f1);break;
    case VIDEO_PAL: fputs("pal", f1);break;
    default: fputs("rgb", f1);break;
    }
    fputc('\n', f1);
}

void config_reset(arguments_t *arguments) {
    memset(arguments, 0, sizeof(arguments_t));
    arguments->vmode = VIDEO_RGB;
    arguments->x = -1;
    arguments->y = -1;
    arguments->filename = NULL;
    arguments->palette_filename = NULL;
    arguments->menu = 1;
}

void configure(int argc, char *argv[], arguments_t *arguments)
{
    int opts = testarg(argc, argv, arguments);
    if (opts < 1) exit(abs(opts));
    if (argc > opts) {
        arguments->filename = (char *)malloc(strlen(argv[opts])+1);
        if (arguments->filename) {
            strcpy(arguments->filename, argv[opts]);
        }
    }
    memcpy(&old_arguments, arguments, sizeof(old_arguments));
    if (arguments->palette_filename) {
        old_arguments.palette_filename = (char *)malloc(strlen(arguments->palette_filename)+1);
        if (old_arguments.palette_filename) {
            strcpy(old_arguments.palette_filename, arguments->palette_filename);
        }
    }
}

void free_config(arguments_t *arguments) {
    free(arguments->filename);
    free(arguments->palette_filename);
    free(old_arguments.palette_filename);
}
