/*
    View64 C64 image viewer
    (c) 1995-2013 Soci/Singular (soci at c64.rulez.org)

    $Id: view64-sdl1.2.c 27 2014-04-06 20:52:33Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include <stdio.h>
#include <string.h>
#ifdef _WIN32
#include <fcntl.h>
#endif

#include "libview64.h"
#include "tvrender.h"
#include "view64_config.h"
#include "main.h"
#include "cfgpath.h"
#include "ini.h"

extern int main2(int, char *[]);

static void word(unsigned char *p, int w) {p[0] = w; p[1] = w >> 8;}
static void dword(unsigned char *p, int w) {word(p, w);word(p + 2, w >> 16);}

static void bmp_save(FILE *f1, const tvrender_image *pixels, unsigned int vmode) {
    int x, y;
    int h = pixels->h * pixels->scale;
    int w = pixels->palette ? pixels->w * pixels->scale : pixels->w;
    unsigned char header[54], *map;
    int with_palette = pixels->palette && pixels->colors <= 256;
    map = (unsigned char *)malloc(w * 3);
    if (map == NULL) {
        return;
    }

    word(header + 0, 0x4D42);
    if (with_palette) {
        if (pixels->colors <= 16) {
            dword(header + 2, (((w+1)/2+3) & ~3)*h + sizeof(header) + pixels->colors * 4);
        } else {
            dword(header + 2, ((w+3) & ~3)*h + sizeof(header) + pixels->colors * 4);
        }
    } else {
        dword(header + 2, ((w*3+3) & ~3)*h + sizeof(header));
    }
    dword(header + 6, 0);
    dword(header + 10, sizeof(header) + pixels->colors * 4);

    dword(header + 14, sizeof(header) - 14);
    dword(header + 18, w);
    dword(header + 22, h);
    word(header + 26, 1);
    if (with_palette) {
        if (pixels->colors <= 16) {
            word(header + 28, 4);
            dword(header + 30, 0);
            dword(header + 34, (((w+1)/2+3) & ~3)*h);
        } else {
            word(header + 28, 8);
            dword(header + 30, 0);
            dword(header + 34, ((w+3) & ~3)*h);
        }
    } else {
        word(header + 28, 24);
        dword(header + 30, 0);
        dword(header + 34, ((w*3+3) & ~3)*h);
    }
    dword(header + 38, 0);
    dword(header + 42, 0);
    dword(header + 46, pixels->colors);
    dword(header + 50, pixels->colors);

    fwrite(header, 1, sizeof(header), f1);
    memset(header, 0, 3);

    if (with_palette) {
        h = pixels->h;
        for (x = 0; x < pixels->colors; x++) {
            const rgb_color *rgb = &pixels->palette[x];
            putc(rgb->b, f1);
            putc(rgb->g, f1);
            putc(rgb->r, f1);
            putc(0, f1);
        }
        if (pixels->colors <= 16) {
            while (h--) {
                unsigned char *p = map, o;
                unsigned char *d = pixels->d + h * pixels->pitch + (pixels->pitch - pixels->w) / 2;
                o = 255;
                for (x = 0; x < pixels->w; x++) {
                    unsigned char b = d[x];
                    int x1;
                    for (x1 = 0; x1 < pixels->scale; x1++) {
                        if (o == 255) {
                            o = b << 4;
                            continue;
                        }
                        *p++ = o | b;
                        o = 255;
                    }
                }
                if (o != 255) {
                    *p++ = o;
                }
                for (y = 0; y < pixels->scale; y++) {
                    fwrite(map, 1, (w + 1) / 2, f1);
                    fwrite(header, 1, ((-((w + 1) / 2) * 3)) & 3, f1);
                }
            }
        } else {
            while (h--) {
                unsigned char *p;
                unsigned char *d = pixels->d + h * pixels->pitch + (pixels->pitch - pixels->w) / 2;
                if (pixels->scale == 1) {
                    fwrite(d, 1, w, f1);
                    fwrite(header, 1, ((-w * 3)) & 3, f1);
                    continue;
                }
                for (x = 0, p = map; x < pixels->w; x++, p += pixels->scale) {
                    memset(p, d[x], pixels->scale);
                }
                for (y = 0; y < pixels->scale; y++) {
                    fwrite(map, 1, w, f1);
                    fwrite(header, 1, ((-w * 3)) & 3, f1);
                }
            }
        }
    } else {
        while (h--) {
            unsigned char *p;
            create_tvpicture_line(pixels, vmode, 0, h, w, map);
            for (x = 0, p = map; x < w; x++, p += 3) {
                unsigned char tmp = p[0];
                p[0] = p[2]; 
                p[2] = tmp; 
            }
            fwrite(map, 1, w * 3, f1);
            fwrite(header, 1, ((-w * 3)) & 3, f1);
        }
    }
    free(map);
}

int main2(int argc, char *argv[])
{
    char cfgdir[MAX_PATH];
    unsigned char *mem;
    view64_screen *rscreen;
    FILE *f1;
    const char *filename = NULL;
    size_t size;
    view64_format format;
    int width, height, zoom;
    rgb_color *palette;
    unsigned int vmode;
    arguments_t arguments;

    config_reset(&arguments);
    get_user_config_file(cfgdir, sizeof(cfgdir), "view64", argv[0]);
    if (cfgdir[0]) {
        FILE *f1 = fopen2(cfgdir, "r");
        if (f1) {
            ini_parse_file(f1, ini_handler, &arguments);
            fclose(f1);
        }
    }
    configure(argc, argv, &arguments);
    filename = arguments.filename;
    width = arguments.width > 4160 ? 4160 : arguments.width;
    height = arguments.height > 2600 ? 2600 : arguments.height;
    vmode = arguments.vmode;
    zoom = arguments.zoom > 13 ? 13 : arguments.zoom;

    if (!filename) {
        puts("Usage: view64pnm [OPTIONS...] PICTURE\n"
             "Try `view64 --help' or `view64 --usage' for more information.\n");
        return EXIT_FAILURE;
    }

    if ((f1 = fopen2(filename, "rb")) == NULL) {
        fprintf(stderr, "Couldn't open file \"%s\"\n", filename);
        free_config(&arguments);
        return EXIT_FAILURE;
    }

    if (fseek(f1, 0, SEEK_END)) {
        size = 65536;
    } else {
        size = ftell(f1);
        if (size < 65536) size = 65536;
    }
    rewind(f1);
    mem = (unsigned char *)calloc(1, size);
    if (mem == NULL) {
        fclose(f1);
        return EXIT_FAILURE;
    }

    size = fread(mem, 1, size, f1);
    fclose(f1);

    format = view64_format_detect(filename, mem, size);
    if (format == VIEW64_UNKNOWN) {
        fprintf(stderr, "%s: Image format not recognized\n", filename);
        free(mem);
        free_config(&arguments);
        return EXIT_FAILURE;
    }
    rscreen = view64_decode(format, mem, size, 0);
    free(mem);
    if (rscreen == NULL) {
        fprintf(stderr, "%s: Can't decode image\n", filename);
        free_config(&arguments);
        return EXIT_FAILURE;
    }

    if (width < 1 || zoom > 0) {
        if (height < 1 || zoom > 0) {
            /* later */
        } else {
            width = width_tvpicture(rscreen->w, vmode, 1) * height / rscreen->h;
        }
    } else if (height < 1) {
        height = (double)rscreen->h * width / width_tvpicture(rscreen->w, vmode, 1);
    } else {
        /* all set */
    }

    if (zoom < 1) {
        zoom = 1;
        while (zoom < 13 && width_tvpicture(rscreen->w, vmode, zoom + 1) <= width && rscreen->h * (zoom + 1) <= height) zoom++;
    }

    f1 = arguments.palette_filename && arguments.palette_filename[0] ? fopen2(arguments.palette_filename, "rb") : NULL;
    palette = load_palette(f1);
    if (f1) fclose(f1);

    f1 = NULL;
    if (arguments.save)
    {
        char *newname = (char *)malloc(strlen(filename) + 5);
        if (newname != NULL) {
            strcpy(newname, filename);
            strcat(newname, ".bmp");
            f1 = fopen2(newname, "wb");
            if (!f1) fprintf(stderr, "Couldn't open file \"%s\"\n", newname);
            free(newname);
        }
    } else {
        f1 = stdout;
#ifdef _WIN32
        setmode(fileno(stdout), O_BINARY);
#endif
    }
    if (f1) {
        tvrender_image *image = create_tvpicture(rscreen, vmode, zoom, palette);
        if (image != NULL) {
            if (arguments.save) bmp_save(f1, image, vmode);
            else {
                int h = image->h * image->scale;
                int w = image->palette ? image->w * image->scale : image->w;
                unsigned char *surface = (unsigned char *)malloc(w * 3);
                if (surface != NULL) {
                    int y;
                    fprintf(f1, "P6\n%d %d\n255\n", w, h);
                    for (y = 0; y < h; y++) {
                        create_tvpicture_line(image, vmode, 0, y, w, surface);
                        fwrite(surface, 1, w * 3, f1);
                    }
                    free(surface);
                } else {
                    fprintf(stderr, "Out of memory error\n");
                }
            }
            free_tv_pixels(image);
        } else {
            fprintf(stderr, "Out of memory error\n");
        }
        if (f1 != stdout) fclose(f1);
        else {
#ifdef _WIN32
            setmode(fileno(stdout), O_TEXT);
#endif
        }
    }
    view64_free_screen(rscreen);
    free(palette);
    free_config(&arguments);
    return (f1 == NULL) ? EXIT_FAILURE : EXIT_SUCCESS;
}
